# -----------------------------------------------------------------------------
# This file is part of icepap (https://github.com/ALBA-Synchrotron/pyIcePAP)
#
# Copyright 2008-2017 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
#
# You should have received a copy of the GNU General Public License
# along with icepap. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# General additions
# TODO: export non official commands as advance commands in new API

# Changes to be supported from the old API
# TODO: export getTime
# TODO: export getSysStatus
# TODO: export getRackStatus
# TODO: export setDefaultConfig
# TODO: export getCurrent
# TODO: export readParameter
# TODO: export writeParameter
# TODO: export isExpertFlagSet
# TODO: export setExpertFlag
# TODO: export getRacksAlive
# TODO: export getDriversAlive
# TODO: export getDecodedStatus
# TODO: export decodeStatus
# TODO: export serr
# TODO: export memory


__all__ = ['IcePAPController']

import time
import logging
from .communication import IcePAPCommunication
from .axis import IcePAPAxis
from .utils import State
from .fwversion import SUPPORTED_VERSIONS, FirmwareVersion


class IcePAPController:
    """
    IcePAP motor controller class.
    """
    ALL_AXES_VALID = set([r * 10 + i for r in range(16) for i in range(1, 9)])

    def __init__(self, host, port=5000, timeout=3, auto_axes=False, **kwargs):
        log_name = '{0}.IcePAPController'.format(__name__)
        self.log = logging.getLogger(log_name)

        self._comm = IcePAPCommunication(host, port, timeout)

        self._aliases = {}
        self._axes = {}

        if auto_axes:
            for axis in self.find_axes(only_alive=True):
                self._axes[axis] = IcePAPAxis(self, axis)

    def __getitem__(self, item):
        if isinstance(item, str):
            item = self._get_axis_for_alias(item)
        if item not in self._axes:
            if item not in self.ALL_AXES_VALID:
                raise ValueError('Bad axis value.')
            self._axes[item] = IcePAPAxis(self, item)
        return self._axes[item]

    def __iter__(self):
        return self._axes.__iter__()

    def __delitem__(self, key):
        self._axes.pop(key)
        aliases_to_remove = []
        for alias, axis in self._aliases.items():
            if key == axis:
                aliases_to_remove.append(alias)
        for alias in aliases_to_remove:
            self._aliases.pop(alias)

    def __repr__(self):
        return '{}({}:{})'.format(type(self).__name__,
                                  self._com.host, self._com.port)

    def __str__(self):
        msg = 'IcePAPController connected ' \
              'to {}:{}'.format(self._com.host, self._com.port)
        return msg

    def _get_axis_for_alias(self, alias):
        if alias not in self._aliases:
            msg = 'There is not any motor with name {0}'.format(alias)
            raise ValueError(msg)
        alias = self._aliases[alias]
        return alias

    def _alias2axisstr(self, alias):
        """
        Method to get the axis motor number. The input can be: string or a
        number or a list with combination of strings and numbers. The result
        depends of the number of inputs.

        :param alias: str or int or [str, int, ...]

        :return: str
        """
        if isinstance(alias, int):
            result = str(alias)
        elif isinstance(alias, str):
            result = str(self._get_axis_for_alias(alias))
        elif isinstance(alias, list):
            result = []
            for i in alias:
                result.append(self._alias2axisstr(i))
            result = ' '.join(result)
        else:
            raise ValueError()
        return result

    def _axesvalues2str(self, axes_values, cast_type=int):
        """
        Method to convert a list of tuples (axis, pos) to a string.

        :param axes_values: [[str/int, float/int]]

        :return: str
        """
        result = ''
        for axis, value in axes_values:
            result += '{0} {1} '.format(self._alias2axisstr(axis),
                                        cast_type(value))
        return result

# -----------------------------------------------------------------------------
#                       Properties
# -----------------------------------------------------------------------------

    @property
    def axes(self):
        """
        Get the axes numbers.

        :return: [int]
        """
        axes = list(self._axes.keys())
        axes.sort()
        return axes

    @property
    def drivers(self):
        """
        Get the drivers IcePAPAxis objects.

        :return: [IcePAPAxis]
        """
        return list(self._axes.values())

    @property
    def ver(self):
        """
        Get the version of the all driver modules: Driver, DSP, FPGA, PCB, IO
        (IcePAP user manual pag. 144).

        :return: dict{module: (ver, date)}
        """
        ans = self.send_cmd('0:?VER INFO')
        return FirmwareVersion(ans)

    @property
    def ver_saved(self):
        """
        Returns the firmware version stored in the master flash memory.

        :return: dict{}
        """
        ans = self.send_cmd('?VER SAVED')
        return FirmwareVersion(ans)

    @property
    def connected(self):
        return self._comm.is_conneted()

    @property
    def mode(self):
        """
        Get the system mode (IcePAP user manual pag. 91).

        :return: str
        """
        return self.send_cmd('?MODE')[0]

    @mode.setter
    def mode(self, value):
        """
        Set the system mode.

        :param value: str
        """
        cmd = 'MODE {0}'.format(value)
        self.send_cmd(cmd)

# -----------------------------------------------------------------------------
#                       Commands
# -----------------------------------------------------------------------------

    def items(self):
        """
        Get the axes and drivers IcePAPAxis objects.

        :return: [(axis, IcePAPAxis),]
        """

        return list(self._axes.items())

    def find_axes(self, only_alive=False):

        # Take the list of racks present in the system
        # IcePAP user manual pag. 137
        racks_present = int(self._comm.send_cmd('?sysstat')[0], 16)
        rack_mask = 1
        axes = []
        for i in range(16):
            if (racks_present & rack_mask << i) > 0:
                # Take the motors presents for a rack.
                cmd = '?sysstat {0}'.format(i)
                drivers_mask = self._comm.send_cmd(cmd)
                # TODO: Analyze if use the present or the alive mask
                if only_alive:
                    # Drivers alive
                    drvs = int(drivers_mask[1], 16)
                else:
                    # Drivers present
                    drvs = int(drivers_mask[0], 16)
                drv_mask = 1
                for j in range(8):
                    if (drvs & drv_mask << j) > 0:
                        axis_nr = i * 10 + j + 1
                        axes.append(axis_nr)
        return axes

    def update_axes(self):
        """
        Method to check if the axes created are presents. In case of no,
        the method will remove the IcePAPAxis object and its aliases.

        Note: The axis can be present but not alive

        :return:
        """
        alive_axes = self.find_axes()
        axes_to_remove = []

        for axis in self._axes:
            if axis not in alive_axes:
                axes_to_remove.append(axis)

        for axis in axes_to_remove:
            self.__delitem__(axis)

    def add_alias(self, alias, axis):
        """
        Set a alias for an axis. The axis can have more than one alias.

        :param alias: str
        :param axis: int
        """
        if axis not in self._axes:
            self._axes[axis] = IcePAPAxis(self, axis)
        self._aliases[alias] = axis

    def add_aliases(self, aliases):
        """
        Set alias for mutiple axes.

        :param aliases: {str: int}
        """
        for alias, axis in aliases.items():
            self.add_alias(alias, axis)

    def get_aliases(self):
        """
        Get the aliases of the system. One axis can have move than one alias.

        :return: {int:[str]}
        """
        aliases = {}
        for key, value in self._aliases.items():
            if value in aliases:
                aliases[value].append(key)
            else:
                aliases[value] = [key]
        return aliases

# -----------------------------------------------------------------------------
#                       IcePAP Commands
# -----------------------------------------------------------------------------
    def send_cmd(self, cmd):
        """
        Communication function used to send any command to the IcePAP
        controller.

        :param cmd: string command following the Programming Interface.

        :return: None or list of string without the command and the CRLF.
        """
        return self._comm.send_cmd(cmd)

    def move(self, axes_pos, group=True, strict=False):
        """
        Start absolute movement for axes motor. The method allows aliases.

        :param axes_pos: [(str/int, int)]
        :param group: bool
        :param strict: bool
        """
        cmd = 'MOVE {0} {1} {2}'.format(['', 'GROUP'][group],
                                        ['', 'STRICT'][strict],
                                        self._axesvalues2str(axes_pos))
        self.send_cmd(cmd)

    def rmove(self, axes_pos, group=True, strict=False):
        """
        Start relative movement for axes motor. The method allows aliases.

        :param axes_pos: [(str/int,int)]
        :param group: bool
        :param strict: bool
        """

        cmd = 'RMOVE {0} {1} {2}'.format(['', 'GROUP'][group],
                                         ['', 'STRICT'][strict],
                                         self._axesvalues2str(axes_pos))
        self.send_cmd(cmd)

    def movep(self, pos, axes, group=True, strict=False):
        """
        Start axes movement to parameter value.

        :param pos: float
        :param axes: [str/int]
        :param group: bool
        :param strict: bool
        """

        cmd = 'MOVEP {0} {1} {2} {3}'.format(['', 'GROUP'][group],
                                             ['', 'STRICT'][strict],
                                             pos,
                                             self._alias2axisstr(axes))
        self.send_cmd(cmd)

    def pmove(self, pos, axes, group=False, strict=False):
        """
        Start parameter movement.

        :param pos: float
        :param axes: [str/int]
        :param group: bool
        :param strict: bool
        """
        cmd = 'PMOVE {0} {1} {2} {3}'.format(['', 'GROUP'][group],
                                             ['', 'STRICT'][strict],
                                             pos,
                                             self._alias2axisstr(axes))
        self.send_cmd(cmd)

    def stop(self, axes):
        """
        Stop multiple axis movement (IcePAP user manual pag. 129).

        :param axes: [str/int]
        """
        cmd = 'STOP {0}'.format(self._alias2axisstr(axes))
        self.send_cmd(cmd)

    def abort(self, axes):
        """
        Abort multiple axis movement (IcePAP user manual pag. 129).

        :param axes: [str/int]
        """
        cmd = 'ABORT {0}'.format(self._alias2axisstr(axes))
        self.send_cmd(cmd)

    def get_fpos(self, axes, register='AXIS'):
        """
        Fast read of multiple positions (IcePAP user manual pag. 73).

        :param axes: [str/int]
        :param register: str
        :return: [float]
        """
        cmd = '?FPOS {0} {1}'.format(register, self._alias2axisstr(axes))
        ans = self.send_cmd(cmd)
        return list(map(float, ans))

    def get_fstatus(self, axes):
        """
        Fast read of multiple status (IcePAP user manual pag. 74).

        :param axes: [str/int]
        :return: [int]
        """
        cmd = '?FSTATUS {0}'.format(self._alias2axisstr(axes))
        ans = self.send_cmd(cmd)
        return [int(i, 16) for i in ans]

    def get_states(self, axes):
        """
        Fast read of the multiples status

        :param axes: [str/int]
        :return: [State]
        """
        fstatus = self.get_fstatus(axes)
        return [State(i) for i in fstatus]

    def get_status(self, axes):
        """
        Read of multiple status (IcePAP user manual pag. 128).

        :param axes: [str/int]
        :return: [int]
        """
        cmd = '?STATUS {0}'.format(self._alias2axisstr(axes))
        ans = self.send_cmd(cmd)
        return [int(i, 16) for i in ans]

    # TODO: optionally compare against saved version.
    def check_version(self):
        """
        Compares the current version installed with the supported
        versions specified in the versions module.

        :return: system version number, -1 if not consistent.
        """
        sys_ver = str(self.ver['SYSTEM']['VER'][0])
        if sys_ver in list(SUPPORTED_VERSIONS.keys()):
            if self.ver.is_supported():
                return self.ver['SYSTEM']['VER'][0]
            else:
                print('Modules versions are not consistent.')
                return -1
        else:
            raise RuntimeError('Version %s not supported' %
                               self.ver['SYSTEM']['VER'][0])

    def reboot(self):
        """
        System reboot (IcePAP user manual pag. 115).
        """
        self.send_cmd('REBOOT')

    def reset(self, rack_nr=None):
        """
        Reset system or rack (IcePAP user manual pag. 117).
        """
        if rack_nr is None:
            rack_nr = ''
        cmd = 'RESET {0}'.format(rack_nr)
        self.send_cmd(cmd)

    def get_rid(self, rack_nrs):
        """
        Get the rack hardware identification string (IcePAP user manual pag.
        118).

        :param rack_nrs: int/[int]
        :return: [str]
        """
        if isinstance(rack_nrs, int):
            rack_nrs = [rack_nrs]
        racks_str = ' '.join(['{}'.format(i) for i in rack_nrs])
        cmd = '?RID {0}'.format(racks_str)
        return self.send_cmd(cmd)

    def get_rtemp(self, rack_nrs):
        """
        Get the rack temperatures (IcePAP user manual pag. 123).

        :param rack_nrs: int/[int]
        :return: [float]
        """
        if isinstance(rack_nrs, int):
            rack_nrs = [rack_nrs]
        racks_str = ' '.join(['{}'.format(i) for i in rack_nrs])
        cmd = '?RTEMP {0}'.format(racks_str)
        return list(map(float, self.send_cmd(cmd)))

    def set_power(self, axes, power_on=True):
        """
        Set axes power state.

        :param axes: [str/int]
        :param power_on: bool
        """
        cmd = 'POWER {0} {1}'.format(['OFF', 'ON'][power_on],
                                     self._alias2axisstr(axes))
        self.send_cmd(cmd)

    def get_power(self, axes):
        """
        Get axes power state.

        :param axes: [str/int]
        :return: [bool]
        """
        cmd = '?POWER {0}'.format(self._alias2axisstr(axes))
        ans = self.send_cmd(cmd)
        return [i.upper() == 'ON' for i in ans]

    def print_commands(self):
        """
        Get the allowed commands (IcePAP user manual pag. 75).
        """
        ans = self.send_cmd('?HELP')
        print('\n'.join(ans))

    def get_pos(self, axes, register='AXIS'):
        """
        Get multiple position.

        :param axes: [str/int]
        :param register: str
        :return: [float]
        """
        cmd = '?POS {0} {1}'.format(register, self._alias2axisstr(axes))
        ans = self.send_cmd(cmd)
        return list(map(int, ans))

    def set_pos(self, axes_pos, register='AXIS'):
        """
        Set multiple positions.

        :param axes_pos: [(str/int,int)]
        :param register: str
        """
        cmd = 'POS {0} {1}'.format(register, self._axesvalues2str(axes_pos))
        self.send_cmd(cmd)

    def get_enc(self, axes, register='AXIS'):
        """
        Get multiple encoder.

        :param axes: [str/int]
        :param register: str
        :return: [float]
        """
        cmd = '?ENC {0} {1}'.format(register, self._alias2axisstr(axes))
        ans = self.send_cmd(cmd)
        return list(map(int, ans))

    def set_enc(self, axes_pos, register='AXIS'):
        """
        Set multiple encoder.

        :param axes_pos: [(str/int,int)]
        :param register: str
        """
        cmd = 'ENC {0} {1}'.format(register, self._axesvalues2str(axes_pos))
        self.send_cmd(cmd)

    def get_homestat(self, axes):
        """
        Get home search status.

        :param axes: [str/int]
        :return: [(str,int)] list of Status, Direction per motor
        """
        cmd = '?HOMESTAT {0}'.format(self._alias2axisstr(axes))
        ans = self.send_cmd(cmd)
        return [(ans[i-1], int(v)) for i, v in enumerate(ans) if i % 2 > 0]

    def get_velocity(self, axes, vtype='NOMINAL'):
        """
        Get multiple velocities (IcePAP user manual pag. 141).

        :param axes: [str/int]
        :return: [float]
        """
        cmd = '?VELOCITY {0} {1}'.format(vtype, self._alias2axisstr(axes))
        ans = self.send_cmd(cmd)
        return list(map(float, ans))

    def set_velocity(self, axes_vel):
        """
        Set multiple velocities (Icepap user manual pag. 141).

        :param axes_vel: [(str/int, float)]
        """
        cmd = 'VELOCITY {0}'.format(self._axesvalues2str(axes_vel,
                                                         cast_type=float))
        self.send_cmd(cmd)

    def get_acctime(self, axes, atype='NOMINAL'):
        """
        Get multiple velocities (IcePAP user manual pag. 141).

        :param axes: [str/int]
        :return: [float]
        """
        cmd = '?ACCTIME {0} {1}'.format(atype, self._alias2axisstr(axes))
        ans = self.send_cmd(cmd)
        return list(map(float, ans))

    def set_acctime(self, axes_acc):
        """
        Set multiple velocities (Icepap user manual pag. 141).

        :param axes_acc: [(str/int, float)]
        """
        cmd = 'ACCTIME {0}'.format(self._axesvalues2str(axes_acc,
                                                        cast_type=float))
        self.send_cmd(cmd)

    def esync(self, axes):
        """
        Synchronize internal position register for multiple axis.

        :param axes: [str/int]
        """
        cmd = 'ESYNC {0}'.format(self._alias2axisstr(axes))
        self.send_cmd(cmd)

    def ctrlrst(self, axes):
        """
        Reset control position register for multiple axis.

        :param axes: [str/int]
        """
        cmd = 'CTRLRST {0}'.format(self._alias2axisstr(axes))
        self.send_cmd(cmd)

    def clear_pmux(self, dest=''):
        """
        Clear the multiplexer configuration. You can pass a destination with
        an optional signal or just the signals to remove (IcePAP user manual,
        page 107).

        :param dest: node to remove
        """
        cmd = 'PMUX REMOVE {0}'.format(dest)
        self.send_cmd(cmd)

    def add_pmux(self, source, dest='', pos=True, aux=True, hard=False):
        """
        Configures a position signal multiplexer configuration (IcePAP user
        manual, page 107).

        :param source: Source node
        :param dest: Target node
        :param pos: Connect the Position signals
        :param aux: Connect the Auxiliary signals
        :param hard: Enabling/Disabling hard flag connection.
        """
        cmd = 'PMUX {0} {1} {2} {3} {4}'.format(['', 'HARD'][hard],
                                                ['', 'POS'][pos],
                                                ['', 'AUX'][aux],
                                                source,
                                                dest)
        self.send_cmd(cmd)

    def get_pmux(self):
        """
        Returns a list of the current signals sources used as axis indexers
        (IcePAP user manual, page 107).

        :return: list of multiplexer configurations.
        """
        return self.send_cmd('?PMUX')

    def sprog(self, component=None, force=False, saving=False):
        """
        Firmware programming command. This command assumes that the firmware
        code will be transferred as a binary data block (IcePAP user manual,
        page 112).

        :param component: { NONE | board adress | DRIVERS | CONTROLLERS| ALL }
        :param force: Force overwrite regardless of being idential versions.
        :param saving: Saves firmware into master board flash.
        """
        force_str = ''
        if component:
            comp_str = str(component).upper()
        else:
            comp_str = 'NONE'
        if force:
            force_str = 'FORCE'
        if not saving:
            save_str = 'NOSAVE'
        else:
            save_str = 'SAVE'
        cmd = '*PROG {} {} {}'.format(comp_str, force_str, save_str)
        self.send_cmd(cmd)

    def prog(self, component, force=False):
        """
        Firmware programming command. This command uses the firmware code
        previously stored in the flash memory of the system master board
        (IcePAP user manual, page 112).

        :param component: { board adress | DRIVERS | CONTROLLERS| ALL }
        :param force: Force overwrite regardless of being idential versions.
        """
        force_str = ''
        if force:
            force_str = 'FORCE'
        prog_str = 'PROG'

        cmd = '{} {} {}'.format(prog_str, str(component).upper(), force_str)
        self.send_cmd(cmd)
        time.sleep(5)

    def get_prog_status(self):
        """
        Request the state of the firmware programing operations (IcePAP user
        manual, page 112).

        :return: { OFF | ACTIVE <progress> | DONE | ERROR }
        """
        try:
            ans = self.send_cmd('?PROG')
        except RuntimeError:
            ans = self.send_cmd('?_PROG')
        return ans

    def disconnect(self):
        """
        Method to close the communication with the IcePAP
        :return:
        """

        self._comm.disconnect()



# -----------------------------------------------------------------------------
# This file is part of icepap (https://github.com/ALBA-Synchrotron/pyIcePAP)
#
# Copyright 2008-2017 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
#
# You should have received a copy of the GNU General Public License
# along with icepap. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import socket
import threading
import struct
import time
import array
import logging

__all__ = ['IcePAPCommunication']

ICEPAP_ENCODING = 'latin-1'


def comm_error_handler(f):
    """
    Error handling function (decorator).

    @param f: target function.
    @return: decorated function with error handling.
    """
    def new_func(*args, **kwargs):
        try:
            ans = f(*args, **kwargs)
            return ans
        except Exception as e:
            msg = 'Problem with the communication. Verify ' \
                  'the hardware. Error: {0}'.format(e)
            raise RuntimeError(msg)
    return new_func


class IcePAPCommunication:
    """
    Class implementing the communication layer for IcePAP motion controller.
    It bases on the socket communication. For the Serial communication see
    manual.
    """
    def __init__(self, host, port=5000, timeout=3):
        self._comm = SocketCom(host=host, port=port, timeout=timeout)

    @property
    def host(self):
        return self._comm.host

    @property
    def port(self):
        return self._comm.port

    @property
    def timout(self):
        return self._comm.timeout

    def send_cmd(self, cmd):
        """
        Method to send commands to the IcePAP controller. It uses acknowledge
        communication (IcePAP User Manual pag. 37).

        :param cmd: Command without acknowledge character and CR and/or LF.
        :return: None or list of string without the command and the CRLF.
        """
        cmd.upper().strip()
        flg_read_cmd = '?' in cmd
        flg_ecamdat_cmd = '*ECAMDAT' in cmd
        flg_listdat_cmd = '*LISTDAT' in cmd
        flg_pardat_cmd = '*PARDAT' in cmd
        use_ack = False
        # The acknowledge character does not have effect on read command.
        # There is a bug on some commands PROG, *PROG, RESET(3.17 does not
        # have problem) and command start by ":"
        bad_cmds = ('PROG', '*PROG', 'RESET', ':')

        if flg_read_cmd or cmd.startswith(bad_cmds) or flg_ecamdat_cmd or \
                flg_listdat_cmd or flg_pardat_cmd:
            cmd = '{0}\r'.format(cmd)
        else:
            cmd = '#{0}\r'.format(cmd)
            use_ack = True

        ans = self._comm.send_cmd(cmd)
        msg = 'Error sending command, IcePAP answered {0}'
        if use_ack:
            if 'OK' not in ans:
                raise RuntimeError(msg.format(ans))
            else:
                result = None
        else:
            if ans is None:
                result = ans
            elif '$' in ans:
                # Multi lines
                ans = ans.split('$')[1]
                lines = ans.split('\n')[1:-1]
                # remove CR
                result = [line.split('\r')[0] for line in lines]
            else:
                ans = ans.split('\r\n')[0]
                result = ans.split()[1:]
                if len(result) == 0:
                    result = None
                elif result[0] == 'ERROR':
                    raise RuntimeError(msg.format(ans))

        return result

    def send_binary(self, ushort_data):
        """
        Method to send a binary data to the IcePAP controller.

        :param ushort_data: Data converted to a unsigned short list.
        """
        self._comm.send_binary(ushort_data=ushort_data)

    def disconnect(self):
        """
        Method to close the communication
        """
        self._comm.disconnect()

    def is_conneted(self):
        return self._comm.connected


# -----------------------------------------------------------------------------
#                           Socket Communication
# -----------------------------------------------------------------------------
class SocketCom:
    """
    Class which implements the Socket communication layer with ASCii interface
    for IcePAP motion controllers.
    """
    def __init__(self, host, port=5000, timeout=3.0):
        log_name = '{0}.SocketCom'.format(__name__)
        self.log = logging.getLogger(log_name)
        self._socket = None
        self._lock = threading.Lock()
        self._stop_thread = False
        self._connect_thread = None
        self._connection_error = ''
        self.host = host
        self.port = port
        self.timeout = timeout
        self.connected = False

        # Start the connection thread
        self._start_thread(wait=False)
        self._connect_thread.join()
        if not self.connected and self._connection_error != '':
            raise RuntimeError(self._connection_error)

    def __del__(self):
        self.disconnect()

    def disconnect(self):
        self._socket.close()
        self.connected = False
        self._stop_thread = True
        self._connect_thread.join()

    @comm_error_handler
    def send_cmd(self, cmd):
        """
        Implementation of the API send command via Socket communication layer.

        :param cmd: string Icepap command.
        :return: Raw string answer for the requested commmand.
        """
        wait_answer = ('#' in cmd) or ('?' in cmd)
        return self._send_data(cmd, wait_answer=wait_answer)

    @comm_error_handler
    def send_binary(self, ushort_data):
        """
        Send data in binary mode.

        :param ushort_data: binary data.
        :return:
        """
        # Prepare Metadata header
        startmark = 0xa5aa555a
        nworddata = len(ushort_data)
        checksum = sum(ushort_data)
        maskedchksum = checksum & 0xffffffff
        data = array.array('H', ushort_data)

        str_startmark = struct.pack('L', startmark)[:4]
        str_nworddata = struct.pack('L', nworddata)[:4]
        str_maskedchksum = struct.pack('L', maskedchksum)[:4]
        str_data = data.tostring()
        str_bin = str_startmark + str_nworddata + str_maskedchksum + str_data
        str_bin += b'\r'
        self._send_data(str_bin, wait_answer=False, encoding=False)

    def _start_thread(self, wait=True):
        self.log.debug('Start thread {0}'.format(self._connect_thread))
        self._connect_thread = threading.Thread(target=self._try_to_connect,
                                                args=[wait])
        self._connect_thread.setDaemon(True)
        self._connect_thread.start()

    def _try_to_connect(self, wait=True):
        self.connected = False
        self._connection_error = ''
        sleep_time = self.timeout / 10.0
        if self._socket is not None:
            try:
                self._socket.close()
            except Exception:
                pass
        while not self._stop_thread:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.settimeout(self.timeout)
            NOLINGER = struct.pack('ii', 1, 0)
            # TODO: protect EBADF [Errno 9] during reboot
            self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER,
                                    NOLINGER)
            try:
                self._socket.connect((self.host, self.port))
                self.connected = True
                break
            except socket.timeout:
                self.log.debug('Fail to connect', exc_info=True)
                if not wait:
                    self._connection_error = \
                        'Timeout error! Check if {} is ON'.format(self.host)
                    break
                time.sleep(sleep_time)
            except OSError as e:
                self._connection_error = \
                    'Fail to connect {}:{}. ' \
                    'Error: {}'.format(self.host, self.port, e.strerror)
                break

    def _send_data(self, data, wait_answer=True, size=8192, encoding=True):
        if encoding:
            raw_data = data.encode(ICEPAP_ENCODING)
        else:
            raw_data = data

        if not self.connected:
            self._start_thread()
            raise RuntimeError('Connection error: No connection with the '
                               'Icepap sytem')

        try:
            with self._lock:
                self.log.debug('RAW_DATA to send: {0}'.format(repr(raw_data)))
                raw_data_size = len(raw_data)
                if raw_data_size > size:
                    # TODO: use python logging
                    self.log.debug('Send multi-lines')
                    n = int(raw_data_size / size)
                    start = 0
                    for i in range(n):
                        end = (i+1) * size
                        self._socket.sendall(raw_data[start:end])
                        start = end
                    if int(raw_data_size % size) > 0:
                        self._socket.sendall(raw_data[start:])
                else:
                    self._socket.sendall(raw_data)
                if wait_answer:
                    answer = self._socket.recv(size)
                    if answer.count(b'$') > 0:
                        # -----------------------------------------------------
                        # WORKAROUND
                        # -----------------------------------------------------
                        # AS IT IS SAID IN
                        # https://docs.python.org/3/howto/sockets.html
                        # SECTION "Using a Socket"
                        #
                        # A protocol like HTTP uses a socket for only one
                        # transfer. The client sends a request, the reads a
                        # reply. That's it. The socket is discarded. This
                        # means that a client can detect the end of the reply
                        # by receiving 0 bytes.
                        #
                        # But if you plan to reuse your socket for further
                        # transfers, you need to realize that there is no
                        # "EOT" (End of Transfer) on a socket. I repeat: if a
                        # socket send or recv returns after handling 0 bytes,
                        # the connection has been broken. If the connection
                        # has not been broken, you may wait on a recv forever,
                        # because the socket will not tell you that there's
                        # nothing more to read (for now). Now if you think
                        # about that a bit, you'll come to realize a
                        # fundamental truth of sockets: messages must either
                        # be fixed length (yuck), or be delimited (shrug), or
                        # indicate how long they are (much better), or end by
                        # shutting down the connection. The choice is entirely
                        # yours, (but some ways are righter than others).
                        #
                        # WE SHOULD WAIT UNTIL THE TERMINATOR CHAR '$' IS
                        # FOUND
                        while answer.count(b'$') < 2:
                            answer = answer + self._socket.recv(size)
                    self.log.debug('RAW_DATA read: {0}'.format(repr(answer)))
                    return answer.decode(ICEPAP_ENCODING)
        except Exception as e:
            self._start_thread()
            raise RuntimeError('Communication error: Error sending command to '
                               'the IcePAP ({0})'.format(e))
